#!/usr/bin/env python3
"""Imports CSV files into an Odoo server by using JSONRPC.
"""
# Version 2.22
import csv
import logging
import sys
from typing import Callable, Dict, List, Tuple, Union

from . import logging_from_verbose, odoo_connect_parser, odoo_login
from .importing import add_importing_file_parsing, extract_info_from_parsed

_logger = logging.getLogger(__name__)

__version__ = "1.0.1"
__date__ = "2020-02-17"
__updated__ = "2020-02-26"


def import_with_jsonrpc(
    login: str,
    password: str,
    database: str,
    protocol: str,
    host: str,
    port: int,
    timeout: int,
    model_filenames: List[Tuple[str, str, str]],
    delimiter: str = ",",
) -> int:
    o, session_id = odoo_login(
        database, host, login, password, port, protocol, timeout
    )

    imd = o.env["ir.model.data"]
    imf = o.env["ir.model.fields"]

    def _ref_split(xmlid: str) -> Tuple[str, str]:
        xmlid_split = xmlid.split(".")
        module, name = xmlid_split if len(xmlid_split) > 1 else ("", xmlid)
        return module, name

    xmlid_cache: Dict[str, int] = dict()

    def _ref(xmlid: str, raise_if_not_found: bool = True) -> int:
        if xmlid not in xmlid_cache:
            full_xmlid = xmlid if "." in xmlid else "." + xmlid
            xmlid_cache[xmlid] = imd.xmlid_to_res_id(
                full_xmlid, raise_if_not_found
            )
        return xmlid_cache[xmlid]

    def _get_imf(model: str, field_name: str):
        imf_ids = imf.search(
            [("model", "=", model), ("name", "=", field_name)]
        )
        return imf.browse(imf_ids)[0]

    def _convert_identity(value):
        return value

    def _convert_identity_nullable(value):
        return value if value else None

    def _convert_bool(value) -> bool:
        return value in (1, "1", "True", "true", True)

    def _convert_int(value) -> int:
        return int(value) if value else 0

    def _convert_float(value) -> float:
        return float(value) if value else 0.0

    def _convert_many2one(xmlid: str) -> Union[None, int]:
        if xmlid:
            return _ref(xmlid)
        return None

    def _convert_many2many(
        xmlids: str
    ) -> List[Tuple[int, int, Union[int, List[int]]]]:
        return (
            [(6, 0, [_ref(xmlid) for xmlid in xmlids.split(",")])]
            if xmlids
            else [(5, 0, 0)]
        )

    for model, csv_file, lang in model_filenames:
        _logger.info(
            "Importing - %s in model %s (lang %s)", csv_file, model, lang
        )
        o.env.context["lang"] = lang
        created: List[int] = list()
        wrote: List[int] = list()
        with open(csv_file, "r") as csvfile:
            reader = csv.DictReader(
                csvfile, delimiter=delimiter, quotechar='"'
            )
            # stored in this dict is the name of the odoo field and
            # a conversion method for data
            headers: Dict[str, Tuple[str, Callable]] = dict()
            for row in reader:
                if not headers:
                    _logger.debug("Reading from headers")
                    for column_name in row.keys():
                        odoo_field_name = column_name
                        if column_name.endswith("/id") or column_name.endswith(
                            ":id"
                        ):
                            odoo_field_name = column_name[:-3]
                        if odoo_field_name != "id":
                            _logger.debug('%s: %s', model, odoo_field_name)
                            field = _get_imf(model, odoo_field_name)
                            ttype = field.ttype
                            _logger.info(
                                "%s.%s type is %s",
                                model,
                                odoo_field_name,
                                ttype,
                            )
                            if ttype == "boolean":
                                converter = _convert_bool
                            elif ttype in (
                                "char",
                                "text",
                                "selection",
                                "date",
                            ):
                                if field.required:
                                    converter = _convert_identity
                                else:
                                    converter = _convert_identity_nullable
                            elif ttype in ("many2one",):
                                converter = _convert_many2one
                            elif ttype in ("many2many",):
                                converter = _convert_many2many
                            elif ttype in ("integer",):
                                converter = _convert_int
                            elif ttype in ("float", "monetary"):
                                converter = _convert_float
                            else:
                                _logger.warning("Unknown field type %s", ttype)
                                converter = _convert_identity
                        else:
                            converter = _convert_identity

                        headers[column_name] = (odoo_field_name, converter)
                    # _logger.debug('Read from headers %s', headers)
                values = dict()
                for column_name in row:
                    odoo_field_name, converter = headers[column_name]
                    values[odoo_field_name] = converter(row[column_name])
                xmlid = values.pop("id")
                write_id = None
                if xmlid:
                    module, name = _ref_split(xmlid)
                    write_id = _ref(xmlid, False)
                if write_id:
                    _logger.debug("Write %s (%d)", model, write_id)
                    o.env[model].write(write_id, values)
                    wrote.append(write_id)
                else:
                    _logger.debug("Creating a %s", model)
                    created_id = o.env[model].create(values)
                    created.append(created_id)
                    if xmlid:
                        _logger.debug(
                            "Adding xmlid %s for %d", xmlid, created_id
                        )
                        imd.create(
                            {
                                "name": name,
                                "module": module,
                                "model": model,
                                "res_id": created_id,
                            }
                        )
                        xmlid_cache[xmlid] = created_id
                if (len(created) + len(model)) % 100 == 0:
                    _logger.info(
                        "%s progress: created %d, wrote %d",
                        model,
                        len(created),
                        len(wrote),
                    )

        _logger.info(
            "%s: created %d, wrote %d", model, len(created), len(wrote)
        )
    return 0


def main(argv=None):  # IGNORE:C0111
    """Parse arguments and launch conversion
    """
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".")[0]
    program_license = """%s

  Created by Vincent Hatakeyama on %s.
  Copyright 2020 XCG Consulting. All rights reserved.

  Licensed under the MIT License

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
""" % (
        program_shortdesc,
        str(__date__),
    )
    parser = odoo_connect_parser(program_license, program_version_message)
    add_importing_file_parsing(parser)
    parser.add_argument(
        "--delimiter", help="CSV delimiter [default: %(default)s]", default=","
    )

    nmspc = parser.parse_args(argv)
    logging_from_verbose(nmspc)

    return import_with_jsonrpc(
        login=nmspc.login,
        password=nmspc.password,
        port=nmspc.port,
        host=nmspc.host,
        protocol=nmspc.protocol,
        timeout=nmspc.timeout,
        database=nmspc.database,
        model_filenames=extract_info_from_parsed(nmspc),
        delimiter=nmspc.delimiter,
    )


if __name__ == "__main__":
    return_code = main(sys.argv[1:])
    if return_code:
        exit(return_code)
